# -*- coding: utf-8 -*-
"""
Utils for 'edx-rbac' module.
"""

import importlib
from collections import defaultdict
from collections.abc import Iterable

from django.apps import apps
from django.conf import settings
from edx_rest_framework_extensions.auth.jwt.authentication import get_decoded_jwt_from_auth
from edx_rest_framework_extensions.auth.jwt.cookies import get_decoded_jwt as get_decoded_jwt_from_cookie
from six import string_types

ALL_ACCESS_CONTEXT = '*'


def request_user_has_implicit_access_via_jwt(decoded_jwt, role_name, context=None):
    """
    Check the request's user access by mapping user's roles found in jwt to local feature roles.

    decoded_jwt is a dict
    role_name is a string
    context is anything

    Returns a boolean.

    Mapping should be in settings and look like:

        SYSTEM_TO_FEATURE_ROLE_MAPPING = {
            'enterprise_admin': ['coupon-management', 'data_api_access'],
            'enterprise_leaner': [],
            'coupon-manager': ['coupon-management']
        }
    """
    if not decoded_jwt:
        return False

    assigned_contexts = contexts_accessible_from_jwt(decoded_jwt, [role_name])
    return _user_has_access(assigned_contexts, context)


def contexts_accessible_from_request(request, role_names):
    """
    Uses the JWT provided from a given `request` to determine
    the set of accessible contexts for the given `role_name`.

    This answers the question: What are all of the contexts accessible to the
    requesting user under the given role via the request's JWT?
    """
    return contexts_accessible_from_jwt(
        get_decoded_jwt(request),
        role_names
    )


def contexts_accessible_from_jwt(decoded_jwt, role_names):
    """
    Given a `decoded_jwt` dictionary and a list of role names,
    returns a set of contexts (identifiers) to which the JWT
    grants access for the given roles.  May contain the "wildcard" `ALL_ACCESS_CONTEXT`,
    which grants access within these roles to any context.
    """
    assigned_contexts_by_feature_role = feature_roles_from_jwt(decoded_jwt)
    accessible_contexts = set()
    for role_name in role_names:
        accessible_contexts.update(
            assigned_contexts_by_feature_role.get(role_name, [])
        )
    return accessible_contexts


def feature_roles_from_jwt(decoded_jwt):
    """
    Get the mapping of feature roles to roles found in the given JWT.

    Given a decoded JWT, returns a mapping of feature role names to list of
    contexts for which that role name applies.  A "context" here usually
    means the primary identifier of some resource.
    """
    jwt_roles_claim = decoded_jwt.get('roles', [])

    feature_roles = defaultdict(list)

    for role_data in jwt_roles_claim:
        # split should be more robust because of our cousekeys having colons
        role_in_jwt, __, context_in_jwt = role_data.partition(':')
        mapped_roles = settings.SYSTEM_TO_FEATURE_ROLE_MAPPING.get(role_in_jwt, [])
        for role in mapped_roles:
            feature_roles[role].append(context_in_jwt)

    return feature_roles


def user_has_access_via_database(user, role_name, role_assignment_class, context=None):
    """
    Check if there is a role assignment for a given user and role.

    The role object itself is found via the role_name. The role_assignment_class's get_context() method can return a
    single context string which could be an ALL_ACCESS_CONTEXT or, incase of multiple user contexts, a list of strings.
    The context argument is evaluated against the context(s) received from the role_assignment_class while accounting
    for the ALL_ACCESS_CONTEXT to grant access.
    """
    if getattr(user, 'is_anonymous', False):
        return False

    assigned_contexts = contexts_accessible_from_database(user, [role_name], role_assignment_class)
    return _user_has_access(assigned_contexts, context)


def contexts_accessible_from_database(user, role_names, role_assignment_class):
    """
    Given a user and role, returns a set of contexts (identifiers) to
    which the user has access for the role.  May contain the "wildcard" `ALL_ACCESS_CONTEXT`,
    which grants access in this role to any context.

    This answers the question: What are all of the contexts accessible to the
    requesting user under the given role via DB-persisted role assignments?
    """
    assigned_contexts = set()

    for _, context in role_assignment_class.get_assignments(user, role_names):
        assigned_contexts.update(
            set_from_collection_or_single_item(context)
        )
    return assigned_contexts


def create_role_auth_claim_for_user(user):
    """
    Create role auth claim for a given user.

    Takes a user, and for each RoleAssignment class specified in config as a
    system wide jwt role associated with that user, creates a list of strings
    denoting the role and context.

    Returns a list.

    This setting is a list of classes whose roles should be added to the
    jwt. The setting should look something like this:

        SYSTEM_WIDE_ROLE_CLASSES = [
            SystemWideConcreteUserRoleAssignment
        ]
    """
    def append_role_auth_claim(role_string, context=None):
        """
        Append the formatted auth claim for a role and context.
        """
        if context:
            contextual_role = '{}:{}'.format(role_string, context)
            role_auth_claim.append(contextual_role)
        else:
            role_auth_claim.append(role_string)

    role_auth_claim = []
    for system_role_loc in settings.SYSTEM_WIDE_ROLE_CLASSES:
        # location can either be a module or a django model
        module_name, func_name = system_role_loc.rsplit('.', 1)
        try:
            # first, assume that this is a plain function
            module = importlib.import_module(module_name)
            role_func = getattr(module, func_name)
        except (ImportError, AttributeError):
            # otherwise, assume that it's a django model
            module = apps.get_model(module_name, func_name)
            role_func = module.get_assignments

        for role_string, context in role_func(user):
            if context:
                if isinstance(context, string_types):
                    append_role_auth_claim(role_string, context)
                else:
                    for item in context:
                        append_role_auth_claim(role_string, item)
            else:
                append_role_auth_claim(role_string)
    return role_auth_claim


def is_iterable(obj):
    """
    Returns True if obj is an instance of collections.abc.Iterable.
    """
    return isinstance(obj, Iterable)


def set_from_collection_or_single_item(obj):
    """
    For iterables that are not strings, returns a set of the iterable.
    For all other types of objects, returns a set containing only the object.
    """
    if is_iterable(obj) and not isinstance(obj, string_types):
        return set(obj)
    return set([obj])


def get_decoded_jwt(request):
    """
    Decodes the request's JWT from either cookies or auth payload and returns it.
    Defaults to an empty dictionary.
    """
    decoded_jwt = get_decoded_jwt_from_cookie(request) or get_decoded_jwt_from_auth(request)
    return decoded_jwt or {}


def has_access_to_all(assigned_contexts):
    """
    Determines whether the `ALL_ACCESS_CONTEXT` token is in the set of assigned contexts.
    """
    return ALL_ACCESS_CONTEXT in assigned_contexts


def _user_has_access(assigned_contexts, requested_context):
    """
    `assigned_contexts` - A set of contexts/identifiers which have been assigned access to some user.
    `requested_context` - 0 or many contexts of which to check access.
    """
    # If there are no contexts assigned, return False.
    if not assigned_contexts:
        return False

    # If the user has access to at least one context, and no particular
    # context was asked to be checked in this call, return True.
    if not requested_context:
        return True

    contexts_to_check = set_from_collection_or_single_item(requested_context)

    return has_access_to_all(assigned_contexts) or contexts_to_check.issubset(assigned_contexts)



# -*- coding: utf-8 -*-
"""
URLs for edx_rbac.
"""

from django.conf.urls import url
from django.views.generic import TemplateView

urlpatterns = [
    url(r'', TemplateView.as_view(template_name="edx_rbac/base.html")),
]



# -*- coding: utf-8 -*-
"""
edx_rbac Django application initialization.
"""

from django.apps import AppConfig


class EdxRbacConfig(AppConfig):
    """
    Configuration for the edx_rbac Django application.
    """

    name = 'edx_rbac'



# -*- coding: utf-8 -*-
"""
Forms to be used for djangoapps extending edx_rbac.
"""

from django import forms
from django.utils.translation import ugettext as _

from edx_rbac.fields import UserFromEmailField


class UserRoleAssignmentAdminForm(forms.ModelForm):
    """
    Custom Form for CRUD operations on models extending UserRoleAssignment.
    """
    user = UserFromEmailField(
        label=_('User Email'),
        required=True
    )

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        instance = kwargs.get('instance')
        initial = kwargs.get('initial', {})
        if instance:
            initial['user'] = instance.user.email
            kwargs['initial'] = initial
        super(UserRoleAssignmentAdminForm, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Database models for edx_rbac.
"""

from django.conf import settings
from django.db import models
from django.db.models.base import ModelBase
from django.db.models.fields import FieldDoesNotExist
from model_utils.models import TimeStampedModel
from six import with_metaclass


class UserRoleAssignmentCreator(ModelBase):
    """
    The model extending UserRoleAssignment should get a foreign key to a model that is a subclass of UserRole.
    """

    def __new__(mcs, name, bases, attrs):    # pylint: disable=arguments-differ
        """
        Override to dynamically create foreign key for objects begotten from abstract class.
        """
        model = super(UserRoleAssignmentCreator, mcs).__new__(mcs, name, bases, attrs)
        if model.__name__ != 'UserRoleAssignment' and 'UserRoleAssignment' in [b.__name__ for b in bases]:
            try:
                model._meta.get_field('role')
            except FieldDoesNotExist:
                if model.role_class and issubclass(model.role_class, UserRole):
                    model.add_to_class(
                        'role',
                        models.ForeignKey(model.role_class, db_index=True, on_delete=models.CASCADE),
                    )
                else:
                    raise Exception('role_class must be defined for any subclass of UserRole!')
        return model


class UserRole(TimeStampedModel):
    """
    Model defining a role a user can have.
    """

    name = models.CharField(unique=True, max_length=255, db_index=True)
    description = models.TextField(null=True, blank=True)

    class Meta:
        """
        Meta class for UserRole.
        """

        abstract = True

    def __str__(self):
        """
        Return human-readable string representation.
        """
        return self.name


class UserRoleAssignment(with_metaclass(UserRoleAssignmentCreator, TimeStampedModel)):
    """
    Model for mapping users and their roles.
    """

    user = models.ForeignKey(settings.AUTH_USER_MODEL, db_index=True, on_delete=models.CASCADE)
    role_class = None

    class Meta:
        """
        Meta class for UserRoleAssignment.
        """

        abstract = True

    def get_context(self):
        """
        Return relevant context data related to this role assignment. Defaults to returning None.
        """
        return None

    @classmethod
    def get_assignments(cls, user, role_names=None):
        """
        Return iterator of (rolename, context).
        """
        if not user.is_anonymous:
            kwargs = {'user': user}
            if role_names:
                kwargs['role__name__in'] = role_names

            for assignment in cls.objects.filter(**kwargs).select_related('role'):
                yield assignment.role.name, assignment.get_context()

    def __str__(self):
        """
        Return human-readable string representation.
        """
        # pylint: disable=no-member
        return '{user}:{role}'.format(
            user=self.user.id,
            role=self.role.name,
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()



# -*- coding: utf-8 -*-
"""
Fields to be used for djangoapps extending edx_rbac.
"""

from django import forms
from django.apps import apps
from django.conf import settings
from django.core.exceptions import ValidationError

USER_MODEL = apps.get_model(settings.AUTH_USER_MODEL)


class UserFromEmailField(forms.EmailField):
    """
    Custom Form Field class for selecting users by entering an email.

    Meant to be used for models with foreign keys to the user table.
    """

    def clean(self, value):
        """
        Override for UserFromEmailField clean method.
        """
        try:
            user = USER_MODEL.objects.get(email=value)
        except USER_MODEL.DoesNotExist:
            raise ValidationError('User with email {} does not exist'.format(value))

        return user



"""
Taken from https://github.com/escodebar/django-rest-framework-rules/blob/master/rest_framework_rules/decorators.py.
"""

import functools

import crum


def permission_required(*permissions, **decorator_kwargs):
    """
    Verify permissions for access to the api.

    :param permissions: Permissions added via django_rules add_perm
    :param decorator_kwargs: Arguments for permission checks
    :return: decorator
    """
    def decorator(view):
        """Verify permissions decorator."""
        @functools.wraps(view)
        def wrapped_view(self, request, *args, **kwargs):
            """Wrap for the view function."""
            fn = decorator_kwargs.get('fn', None)
            if callable(fn):
                obj = fn(request, *args, **kwargs)
            else:
                obj = fn

            crum.set_current_request(request)

            missing_permissions = [perm for perm in permissions
                                   if not request.user.has_perm(perm, obj)]
            if any(missing_permissions):
                # raises a permission denied exception causing a 403 response
                self.permission_denied(
                    request,
                    message=(u'Missing: {}'
                             .format(', '.join(missing_permissions)))
                )

            return view(self, request, *args, **kwargs)
        return wrapped_view

    return decorator



# -*- coding: utf-8 -*-
"""
Mixin adapted from https://github.com/escodebar/django-rest-framework-rules/blob/master/rest_framework_rules/mixins.py.

Keeps py2 backward compatibility and only holds on to the necessary bits of the mixin needed.
"""

import crum
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property
from six import string_types

from edx_rbac import utils


class PermissionRequiredMixin:
    """
    Mixin for checking permissions with django rules.
    """

    object_permission_required = None
    permission_required = None

    def get_permission_required(self):
        """
        Return permissions required for the view it is mixed into.
        """
        if self.permission_required is None:
            # This prevents a misconfiguration of the view into which the mixin
            # is mixed. If the mixin is used, at least one permission should be
            # required.
            raise ImproperlyConfigured(
                '{0} is missing the permission_required attribute. Define '
                '{0}.permission_required, or override '
                '{0}.get_permission_required().'
                .format(self.__class__.__name__)
            )
        if isinstance(self.permission_required, string_types):
            perms = (self.permission_required, )
        else:
            perms = self.permission_required
        return perms

    def check_permissions(self, request):
        """
        Check through permissions required and throws a permission_denied if missing any.

        If `get_permission_object` is implemented, it will be called and should return the object
        for which the @rules predicate checks against.
        """
        crum.set_current_request(request)
        user = request.user

        if hasattr(self, 'get_permission_object') and callable(self.get_permission_object):
            obj = self.get_permission_object()
        else:
            obj = None

        missing_permissions = [perm for perm in self.get_permission_required()
                               if not user.has_perm(perm, obj)]

        if any(missing_permissions):
            self.permission_denied(
                request,
                message=('MISSING: {}'.format(', '.join(missing_permissions))))


class PermissionRequiredForListingMixin(PermissionRequiredMixin):
    """
    Mixin for checking role-based access against multiple objects, specifically
    using the "list" action of a DRF ViewSet.  The brunt of the work is done in
    the `accessible_contexts` cached property - this property determines the set
    of "contexts" (probably identifiers of some model) to which the requesting
    user has access under any of the `allowed_roles`.  It's pretty opinionated and
    you _must_ define the following fields:

    `list_lookup_field` (string) - The field of the model you want to list instances of
    that holds values defined by `accessible_contexts`.

    `allowed_roles` (list) - A list of one or more roles against which access is checked.  It's disjunctive,
    which means that if "role_1" grants access to resources "A" and "B", and
    "role_2" grants access to resource "C", defining `allowed_roles = ['role_1', 'role_2']`
    will cause resources "A", "B", and "C" to be returned by a listing operation
    for a user who is granted both "role_1" and "role_2".

    `role_assignment_class` (class) - The RoleAssignmentClass against which DB-defined access is checked.

    `base_queryset` (property) - A queryset which acts as the "base case".  It should generally return
    all accessible instances for a user who has access to anything within this viewset (like a superuser
    or admin).
    """
    # This flag indicates whether requesting users with `is_staff = True`
    # should ever encounter a PermissionDenied exception.
    staff_are_never_forbidden = True

    # This flag indicates whether requesting users with `is_superuser = True`
    # should always have the `ALL_ACCESS_CONTEXT` included in their set of accessible contexts.
    superusers_can_access_anything = True

    # The field (as a string) of the model you want to list instances of.
    list_lookup_field = None

    # A list of one or more roles against which access is checked.
    allowed_roles = []

    # The RoleAssignmentClass against which DB-defined access is checked
    role_assignment_class = None

    @property
    def base_queryset(self):
        """
        Should return a queryset that acts as the "base case".
        It should generally return all accessible instances for a user who has access
        to everything within this viewset (like a superuser or admin).
        """
        raise NotImplementedError

    @cached_property
    def request_action(self):
        """
        Returns `self.action` if it exists, None otherwise.
        """
        return getattr(self, 'action', None)

    @cached_property
    def accessible_contexts(self):
        """
        Cached set of contexts (usually model identifiers) the requesting user has access to.
        Returns a set that contains the `ALL_ACCESS_CONTEXT` identifier
        if the requesting user is a superuser.
        """
        contexts_via_jwt = utils.contexts_accessible_from_request(self.request, self.allowed_roles)
        contexts_via_db = set()

        if self.role_assignment_class:
            contexts_via_db = utils.contexts_accessible_from_database(
                self.request.user, self.allowed_roles, self.role_assignment_class
            )

        if self.request.user.is_superuser and self.superusers_can_access_anything:
            contexts_via_db.add(utils.ALL_ACCESS_CONTEXT)

        return contexts_via_jwt | contexts_via_db

    def check_permissions(self, request):
        """
        If dealing with a "list" action, goes through some customized
        logic to check which contexts are accessible by the requesting
        user.  If none are, and the user is not staff/super, raise
        a `PermissionDenied` exception.  Uses the parent class's `check_permissions()`
        method if the request action is not "list".
        """
        if self.request_action == 'list':
            # Super-users and staff won't get Forbidden responses,
            # but depending on their assigned roles, staff may
            # get an empty result set.
            if request.user.is_superuser:
                return
            if request.user.is_staff and self.staff_are_never_forbidden:
                return
            if not self.accessible_contexts:
                self.permission_denied(request)
        else:
            super().check_permissions(request)

    def get_queryset(self):
        """
        Expects `self.base_queryset` to be explicitly defined as the "base case"
        and `self.list_lookup_field` to be defined.
        You'll most likely want it to be `MyModel.objects.all().order_by('some_field')`.
        """
        if getattr(self, 'base_queryset', None) is None:
            raise Exception('{} must have a non-null "base_queryset" field.'.format(self.__class__))
        if not getattr(self, 'list_lookup_field', None):
            raise Exception('{} must have a truthy "list_lookup_field" field.'.format(self.__class__))

        if self.request_action == 'list':
            if not self.accessible_contexts:
                return self.base_queryset.none()
            if utils.has_access_to_all(self.accessible_contexts):
                return self.base_queryset
            kwargs = {self.list_lookup_field + '__in': self.accessible_contexts}
            return self.base_queryset.filter(**kwargs)

        return self.base_queryset



# -*- coding: utf-8 -*-
"""
Django admin integration for djangoapps using edx-rbac.
"""

from django.contrib import admin

from edx_rbac.admin.forms import UserRoleAssignmentAdminForm


class UserRoleAssignmentAdmin(admin.ModelAdmin):
    """
    Django admin for UserRoleAssignment.
    """

    class Meta:
        """
        Meta class for UserRoleAssignmentAdmin.
        """
        abstract = True

    list_display = (
        'user', 'role'
    )

    list_filter = ('role',)
    search_fields = ('user__email', 'role__name')
    fields = ('user', 'role',)
    form = UserRoleAssignmentAdminForm
